<?php

namespace Drupal\field_tools\Form;

use Drupal\Core\Field\FieldStorageDefinitionInterface;
use Drupal\Core\Entity\EntityTypeBundleInfoInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Serialization\Yaml;
use Drupal\field_tools\FieldCloner;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a form to import multiple fields to base field code.
 *
 * This allow fields initially created in the admin UI to be converted to base
 * fields.
 *
 * @todo Use ConfirmFormBase.
 */
class ConfigMultipleImportForm extends FormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'field_tools_config_multiple_import_form';
  }

   /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form['import'] = [
      '#title' => $this->t('Paste your configuration here. WARNING: This currently only works with field and field storage configs exported from a bundle field tools config export.'),
      '#type' => 'textarea',
      '#rows' => 24,
      '#required' => TRUE,
    ];

    $form['actions'] = ['#type' => 'actions'];
    $form['actions']['submit'] = [
      '#type' => 'submit',
      '#value' => $this->t('Import'),
      '#button_type' => 'primary',
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $import = $form_state->getValue('import');

    // Split up into different YAMLs.
    $data = explode("\n", $import);
    $yamls = [];
    $current = [];
    // ARGH this is really ugly! Refactor!
    foreach ($data as $line) {
      // Start a fresh YAML and take the config prefix.
      if (str_starts_with($line, '_config_prefix')) {
        if ($current) {
          $yamls[] = $current;
        }

        $current = [];
        $current['config_prefix'] = trim(explode(':', $line)[1]); // !! todo
        continue;
      }

      $current['yaml'][] = $line;
    }
    // Add the last one.
    $yamls[] = $current;

    foreach ($yamls as $yaml) {
      $config_item_data = Yaml::decode(implode("\n", $yaml['yaml']));
      // @todo Handle errors.
      //
      // HACK! Assume it's a field or a field storage, because ConfigImporter
      // is a pain to work with and I CBA.
      // See https://www.drupal.org/project/drupal/issues/3123491.
      // @todo Make this actually import config, not save an entity!
      $entity_type_id = match($yaml['config_prefix']) {
        'field.storage' => 'field_storage_config',
        'field.field' => 'field_config',
      };

      $entity = \Drupal::service('entity_type.manager')->getStorage($entity_type_id)->create($config_item_data);
      $entity->save();
    }
  }

}
